@@ -2,6 +2,7 @@
 
 from __future__ import annotations
 
+import copy
 import fnmatch
 import logging
 import os
@@ -107,6 +108,9 @@ def extract_repos(
             if "name" not in conf:
                 conf["name"] = repo
 
+            if "workspace_root" not in conf:
+                conf["workspace_root"] = directory
+
             if "path" not in conf:
                 conf["path"] = expand_dir(
                     pathlib.Path(expand_dir(pathlib.Path(directory), cwd=cwd))
@@ -283,13 +287,10 @@ def load_configs(
     return repos
 
 
-ConfigDictTuple = tuple["ConfigDict", "ConfigDict"]
-
-
 def detect_duplicate_repos(
     config1: list[ConfigDict],
     config2: list[ConfigDict],
-) -> list[ConfigDictTuple]:
+) -> list[tuple[ConfigDict, ConfigDict]]:
     """Return duplicate repos dict if repo_dir same and vcs different.
 
     Parameters
@@ -300,13 +301,13 @@ def detect_duplicate_repos(
 
     Returns
     -------
-    list[ConfigDictTuple]
+    list[tuple[ConfigDict, ConfigDict]]
         List of duplicate tuples
     """
     if not config1:
         return []
 
-    dupes: list[ConfigDictTuple] = []
+    dupes: list[tuple[ConfigDict, ConfigDict]] = []
 
     repo_dirs = {
         pathlib.Path(repo["path"]).parent / repo["name"]: repo for repo in config1
@@ -424,3 +425,121 @@ def is_config_file(
         extensions = [".yml", ".yaml", ".json"]
     extensions = [extensions] if isinstance(extensions, str) else extensions
     return any(filename.endswith(e) for e in extensions)
+
+
+def save_config_yaml(config_file_path: pathlib.Path, data: dict[t.Any, t.Any]) -> None:
+    """Save configuration data to a YAML file.
+
+    Parameters
+    ----------
+    config_file_path : pathlib.Path
+        Path to the configuration file to write
+    data : dict
+        Configuration data to save
+    """
+    yaml_content = ConfigReader._dump(
+        fmt="yaml",
+        content=data,
+        indent=2,
+    )
+    config_file_path.write_text(yaml_content, encoding="utf-8")
+
+
+def canonicalize_workspace_path(
+    label: str,
+    *,
+    cwd: pathlib.Path | None = None,
+) -> pathlib.Path:
+    """Convert a workspace root label to an absolute canonical path."""
+    cwd = cwd or pathlib.Path.cwd()
+    label_path = pathlib.Path(label)
+    return expand_dir(label_path, cwd=cwd)
+
+
+def workspace_root_label(
+    workspace_path: pathlib.Path,
+    *,
+    cwd: pathlib.Path | None = None,
+    home: pathlib.Path | None = None,
+) -> str:
+    """Create a normalized label for a workspace root path."""
+    cwd = cwd or pathlib.Path.cwd()
+    home = home or pathlib.Path.home()
+
+    if workspace_path == cwd:
+        return "./"
+
+    try:
+        relative_to_home = workspace_path.relative_to(home)
+        label = f"~/{relative_to_home.as_posix()}"
+    except ValueError:
+        label = workspace_path.as_posix()
+
+    if label != "./" and not label.endswith("/"):
+        label += "/"
+
+    return label
+
+
+def normalize_workspace_roots(
+    config_data: dict[str, t.Any],
+    *,
+    cwd: pathlib.Path | None = None,
+    home: pathlib.Path | None = None,
+) -> tuple[dict[str, t.Any], dict[pathlib.Path, str], list[str], int]:
+    """Normalize workspace root labels and merge duplicate sections."""
+    cwd = cwd or pathlib.Path.cwd()
+    home = home or pathlib.Path.home()
+
+    normalized: dict[str, t.Any] = {}
+    path_to_label: dict[pathlib.Path, str] = {}
+    conflicts: list[str] = []
+    change_count = 0
+
+    for label, value in config_data.items():
+        canonical_path = canonicalize_workspace_path(label, cwd=cwd)
+        normalized_label = path_to_label.get(canonical_path)
+
+        if normalized_label is None:
+            normalized_label = workspace_root_label(
+                canonical_path,
+                cwd=cwd,
+                home=home,
+            )
+            path_to_label[canonical_path] = normalized_label
+
+            if isinstance(value, dict):
+                normalized[normalized_label] = copy.deepcopy(value)
+            else:
+                normalized[normalized_label] = value
+
+            if normalized_label != label:
+                change_count += 1
+        else:
+            change_count += 1
+            existing_value = normalized.get(normalized_label)
+
+            if isinstance(existing_value, dict) and isinstance(value, dict):
+                for repo_name, repo_config in value.items():
+                    if repo_name not in existing_value:
+                        existing_value[repo_name] = copy.deepcopy(repo_config)
+                        change_count += 1
+                    elif existing_value[repo_name] != repo_config:
+                        conflict_message = (
+                            "Workspace root '{label}' contains conflicting definitions "
+                            "for repository '{repo}'. Keeping the existing entry."
+                        )
+                        conflicts.append(
+                            conflict_message.format(
+                                label=normalized_label,
+                                repo=repo_name,
+                            )
+                        )
+            elif existing_value != value:
+                conflict_message = (
+                    "Workspace root '{label}' contains conflicting non-dictionary "
+                    "values. Keeping the existing entry."
+                )
+                conflicts.append(conflict_message.format(label=normalized_label))
+
+    return normalized, path_to_label, conflicts, change_count
